.wd-table {
  overflow: hidden;
}

.wd-table-scroller {
  overflow: auto;
  max-width: 100%;
  background: var(--wd-color-surface);
}

.wd-table-scroller table {
  border-collapse: collapse;
  width: 100%;
  min-width: 100%;
  background: var(--wd-color-surface);
}

.wd-table-scroller th,
.wd-table-scroller td {
  border-bottom: 1px solid var(--wd-border-muted);
  padding: 0.75rem 1rem;
  text-align: left;
  font-size: 0.95rem;
  color: var(--wd-text-secondary);
}

.wd-table-scroller thead th {
  font-weight: 600;
  color: var(--wd-text-primary);
  background: var(--wd-color-surface);
}

.wd-table-scroller tbody tr:nth-child(even) {
  background: var(--wd-panel-muted);
}

.wd-table-scroller tbody tr:hover {
  background: var(--wd-accent-soft);
}

.wd-table-scroller tbody tr:last-child td {
  border-bottom: none;
}

.wd-table-mini {
  position: sticky;
  top: calc(var(--wd-top-offset, 0px));
  z-index: 2;
  display: flex;
  gap: 0.75rem;
  align-items: baseline;
  padding: 0.5rem 1rem;
  margin-bottom: 0.25rem;
  background: color-mix(in srgb, var(--wd-color-surface) 90%, var(--wd-border-muted) 10%);
  border-bottom: 1px solid var(--wd-border-muted);
  font-size: 0.75rem;
  line-height: 1.1;
  color: var(--wd-text-muted);
}

.wd-table-mini[hidden] {
  display: none;
}

.wd-table-mini-label {
  white-space: nowrap;
  text-transform: uppercase;
  letter-spacing: 0.04em;
}
